from .base import BaseOperation
import pandas as pd
from pyspark.sql import DataFrame as SparkDataFrame
from pyrecdp.core.utils import *
import copy

class CategorifyOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.feature_in_out = self.op.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = False
    
    def get_function_pd(self):
        feature_in_out = copy.deepcopy(self.feature_in_out)
        def categorify(df):
            from sklearn.preprocessing import LabelEncoder
            for feature, (dict_path, feature_out) in feature_in_out.items():
                encoder = LabelEncoder()
                encoder = get_encoder_np(encoder, dict_path)
                encoder.fit(df[feature])
                save_encoder_np(encoder, dict_path)
                df[f"{feature_out}"] = pd.Series(encoder.transform(df[feature]))
                
            return df
        return categorify

class GroupedCategorifyOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.feature_in_out = self.op.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = False

    def get_function_pd(self):
        feature_in_out = copy.deepcopy(self.feature_in_out)
        def group_categorify(df, feature_name, grouped_features, dict_path = None):
            encoder = get_encoder_df(dict_path)
            if isinstance(encoder, type(None)):
                k = [i for i in df.columns if i not in grouped_features]
                if len(k) == 0:
                    return ret, None, 0
                k = k[0]
                encoder = df.groupby(by = grouped_features, as_index = False)[k].count().drop(k, axis = 1)
                encoder[feature_name] = encoder.index
                save_encoder_df(encoder, dict_path)
            ret = df.merge(encoder, on = grouped_features, how = 'left')
            return ret
        def categorify(df):
            for feature_out, feature in feature_in_out.items():
                dict_path = None
                df = group_categorify(df, feature_out, feature, dict_path)
            return df
        return categorify


from .binned import BinnedFeatureGenerator
from .category import CategoryFeatureGenerator
from .datetime import DatetimeFeatureGenerator
from .drop import DropUselessFeatureGenerator
from .name import RenameFeatureGenerator
from .fillna import FillNaFeatureGenerator
from .type import TypeCheckFeatureGenerator,TypeConvertFeatureGenerator
from .nlp import DecodedTextFeatureGenerator, TextFeatureGenerator
from .geograph import GeoFeatureGenerator, CoordinatesInferFeatureGenerator
from .relation import RelationalFeatureGenerator
from .encode import OneHotFeatureGenerator, ListOneHotFeatureGenerator, TargetEncodeFeatureGenerator, LabelEncodeFeatureGenerator
from .feature_transform import ConvertToNumberFeatureGenerator

relation_builder_list = [
    RelationalFeatureGenerator
]

profiler_feature_generator_list = [
    ConvertToNumberFeatureGenerator,
    TypeConvertFeatureGenerator,
]

label_feature_generator_list = [
    RenameFeatureGenerator,
    TypeConvertFeatureGenerator,
    FillNaFeatureGenerator,
    LabelEncodeFeatureGenerator,
]

pre_feature_generator_list = [
    CoordinatesInferFeatureGenerator,
    ConvertToNumberFeatureGenerator,
    TypeConvertFeatureGenerator,
    FillNaFeatureGenerator,
    RenameFeatureGenerator,
]

transformation_generator_list = [
    DecodedTextFeatureGenerator,
    DatetimeFeatureGenerator,
    GeoFeatureGenerator,
    TextFeatureGenerator,
]

local_encode_generator_list = [
    OneHotFeatureGenerator,
    ListOneHotFeatureGenerator,
]

pre_enocode_feature_generator_list = [
    DropUselessFeatureGenerator,
]

global_dict_index_generator_list = [
    BinnedFeatureGenerator,
    CategoryFeatureGenerator,
    #TargetEncodeFeatureGenerator
]

post_feature_generator_list = [
    RenameFeatureGenerator
]

final_generator_list = [
    TypeCheckFeatureGenerator,
    DropUselessFeatureGenerator,
]


from pyrecdp.core import SeriesSchema
from pyrecdp.primitives.operations import Operation
import pandas as pd
import numpy as np
from pandas.api import types as pdt
import copy
from featuretools.primitives.base import TransformPrimitive
from tqdm import tqdm

def try_category(s):
    if pdt.is_categorical_dtype(s) and not pdt.is_bool_dtype(s):
        return False
    n_unique = s.nunique()
    total_len = len(s)
    threshold = (total_len / 5)
    if 1 <= n_unique <= threshold:
        return True     
    return False

def try_onehot(s):
    n_unique = s.nunique()
    if n_unique < 10:
        return s.unique().tolist()
    return False

def try_datetime(s):
    if pdt.is_datetime64_any_dtype(s):
        return False
    if not pdt.is_string_dtype(s):
        return False
    if s.isnull().all():
        return False
    try:
        if len(s) > 500:
            # Sample to speed-up type inference
            result = s.sample(n=500, random_state=0)
        result = pd.to_datetime(result, errors='coerce')
        if result.isnull().mean() > 0.8:  # If over 80% of the rows are NaN
            return False
        else:
            return True
    except:
        return False

def try_re_numeric(s):
    if s.isnull().all():
        return True
    try:
        if len(s) > 500:
            # Sample to speed-up type inference
            result = s.sample(n=500, random_state=0)
        import re
        result = result.apply(lambda x: "".join(re.findall(r'\d+.*\d*', x)))
        result = pd.to_numeric(result, errors='coerce')
        if result.isnull().mean() <= 0.8:  # If over 80% of the rows are NaN
            return True
        return False
    except:
        return False
     
def try_numeric(s):
    if pdt.is_numeric_dtype(s):
        return True
    if s.isnull().all():
        return True
    try:
        if len(s) > 500:
            # Sample to speed-up type inference
            result = s.sample(n=500, random_state=0)
        result = pd.to_numeric(result, errors='coerce')
        if result.isnull().mean() <= 0.8:  # If over 80% of the rows are NaN
            return True
        return False
    except:
        return False

def try_list_string(s):
    if not pdt.is_string_dtype(s):
        return False
    result = s[s.notna()]
    try:
        result = pd.Series(s).str.split(';')
        avg_words = result.str.len().mean()
        if avg_words > 1:
            estimated_size = result.explode().nunique()
            if estimated_size < 100:
                return (';', result.explode().unique().tolist())
    except:
        pass
    try:
        result = pd.Series(s).str.split(',')
        avg_words = result.str.len().mean()
        if avg_words > 1:
            estimated_size = result.explode().nunique()
            if estimated_size < 100:
                return (',', result.explode().unique().tolist())
    except:
        pass
    
    return False
         
class TypeInferFeatureGenerator():
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
   
    def fit_prepare(self, pipeline, children, max_idx, df):
        ret_pa_fields = []
        pa_schema = pipeline[children[0]].output
        for pa_field, feature_name in tqdm(zip(pa_schema, df.columns), total=len(df.columns)):
            config = {}
            if try_datetime(df[feature_name]):
                config['is_datetime'] = True
            if try_category(df[feature_name]):
                config['is_categorical'] = True
            if try_numeric(df[feature_name]):
                config['is_numeric'] = True
            elif try_re_numeric(df[feature_name]):
                config['is_re_numeric'] = True
            config['is_list_string'] = try_list_string(df[feature_name])
            if config['is_list_string'] is False:
                skip = ('is_numeric' in config and config['is_numeric']) or ('is_re_numeric' in config and config['is_re_numeric'])
                if not skip:
                    config['is_onehot'] = try_onehot(df[feature_name])
            
            pa_field.copy_config_from(config)
            ret_pa_fields.append(pa_field)
        
        # append to pipeline
        cur_idx = max_idx + 1
        config = [x.mydump() for x in ret_pa_fields]
        pipeline[cur_idx] = Operation(cur_idx, children, ret_pa_fields, op = 'type_infer', config = config)
        return pipeline, cur_idx, cur_idx
    



import logging
from pyrecdp.core.utils import Timer, infer_problem_type
from pyrecdp.core.dataframe import DataFrameAPI

from pyrecdp.autofe import FeatureWrangler, FeatureProfiler, RelationalBuilder, FeatureEstimator

import os

logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.ERROR, datefmt='%I:%M:%S')
logger = logging.getLogger(__name__)

class AutoFE():
    def __init__(self, dataset, label, *args, **kwargs):
        self.label = label
        self.auto_pipeline = {'relational': None, 'profiler': None, 'wrangler': None, 'estimator': None}
        if isinstance(dataset, dict):
            self.auto_pipeline['relational'] = RelationalBuilder(dataset=dataset, label=label)
        else:
            print("AutoFE started to profile data")
            self.auto_pipeline['profiler'] = FeatureProfiler(dataset=dataset, label=label)
            print("AutoFE started to create data pipeline")
            self.auto_pipeline['wrangler'] = FeatureWrangler(dataset=dataset, label=label)
            pipeline = self.auto_pipeline['wrangler']
            config = {
                'model_file': 'autofe_lightgbm.mdl',
                'objective': infer_problem_type(pipeline.dataset[pipeline.main_table][label]),
                'model_name': 'lightgbm'}
            self.auto_pipeline['estimator'] = FeatureEstimator(data_pipeline = pipeline, config = config)

    def fit_transform(self, engine_type = 'pandas', no_cache = False, *args, **kwargs):
        print("AutoFE started to execute data")
        ret_df = None
        if self.auto_pipeline['relational']:
            pipeline = self.auto_pipeline['relational']
            ret_df = pipeline.fit_transform(engine_type, data = ret_df)
            print("AutoFE started to profile data")

        if self.auto_pipeline['estimator']:
            pipeline = self.auto_pipeline['estimator']
            ret_df = pipeline.fit_transform(engine_type, data = ret_df)
        return ret_df
    
    def profile(self, engine_type = 'pandas'):
        if isinstance(self.auto_pipeline['profiler'], type(None)):
            return "feature profile support for multiple table is WIP"
        return self.auto_pipeline['profiler'].visualize_analyze(engine_type)

    def feature_importance(self):
        import pandas as pd
        if isinstance(self.auto_pipeline['estimator'], type(None)):
            return "feature estimator support for multiple table is WIP"
        fe_imp_dict = self.auto_pipeline['estimator'].get_feature_importance()
        feat_importances = pd.Series([i[1] for i in fe_imp_dict], [i[0] for i in fe_imp_dict])
        feat_importances = feat_importances[feat_importances > 0].sort_values()
        height = int(len(feat_importances) * 0.5)
        return feat_importances.plot(kind='barh', figsize=(15,height))

    def plot(self):
        if self.auto_pipeline['relational']:
            return self.auto_pipeline['relational'].plot()
        else:
            return self.auto_pipeline['estimator'].plot()
        
    def get_transformed_data(self):
        if self.auto_pipeline['relational']:
            return self.auto_pipeline['relational'].get_transformed_cache()
        else:
            ret_df = self.auto_pipeline['estimator'].get_transformed_cache()
            return ret_df
        
    def get_feature_list(self):
        fe_imp_dict = self.auto_pipeline['estimator'].get_feature_importance()
        feature_list = [i[0] for i in fe_imp_dict if i[1] > 0]
        return feature_list
    
    def add_operation(self, config):
        if self.auto_pipeline['relational']:
            return self.auto_pipeline['relational'].add_operation(config)
        else:
            return self.auto_pipeline['estimator'].add_operation(config)
        
    def delete_operation(self, idx):
        if self.auto_pipeline['relational']:
            return self.auto_pipeline['relational'].delete_operation(idx)
        else:
            return self.auto_pipeline['estimator'].delete_operation(idx)


import pandas as pd

class DataFrameAPI:
    @staticmethod
    def instiate(df):
        if isinstance(df, pd.DataFrame):
            from pandas_flavor import register_dataframe_method
            @register_dataframe_method
            def may_sample(df, nrows = 100000):
                if df.shape[0] > nrows:
                    return df.sample(n=nrows, random_state = 123)
                else:
                    return df
            return df
        elif isinstance(df, str):
            # open the path depends on the size
            raise NotImplementedError("Instiate DataFrame based on path is not implemented yet.")
        raise NotImplementedError("Instiate DataFrame based on non-pandas dataframe is not implemented yet.")


from .base import BaseOperation
import pandas as pd
from pyspark.sql import DataFrame as SparkDataFrame
import copy
import numpy as np
from pyrecdp.core.utils import *
from IPython.display import display

class OnehotEncodeOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.config = self.op.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = False
    
    def get_function_pd(self):
        config = copy.deepcopy(self.config)
        def encode(df):
            for feature, keys in config.items():
                selected_columns = [f"{feature}__{key}" for key in keys]
                one_hot_df = pd.get_dummies(df[feature], prefix = f"{feature}_")
                one_hot_df = one_hot_df.loc[:, one_hot_df.columns.isin(selected_columns)]
                df = pd.concat([df, one_hot_df], axis=1)
            return df
        return encode
    
class ListOnehotEncodeOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.config = self.op.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = False
    
    def get_function_pd(self):
        config = copy.deepcopy(self.config)
        def encode(df):
            from sklearn.preprocessing import MultiLabelBinarizer
            
            for feature, (sep, keys) in config.items():
                splitted_s = df[feature].str.split(sep).apply(lambda x: [i for i in x if i is not ""])
                mlb = MultiLabelBinarizer()
                encoded = mlb.fit_transform(splitted_s)
                names = [f"{feature}_{key}" for key in mlb.classes_]
                selected_columns = [f"{feature}_{key}" for key in keys]
                one_hot_df = pd.DataFrame(encoded, columns=names)
                one_hot_df = one_hot_df.loc[:, one_hot_df.columns.isin(selected_columns)]
                df = pd.concat([df, one_hot_df], axis=1)
            return df
        return encode
    

class TargetEncodeOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.feature_in = self.op.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = True
    
    def get_function_pd(self):
        from category_encoders import TargetEncoder
        feature_in = copy.deepcopy(self.feature_in)
        def encode(df):
            added_data = {}
            for feature, (dict_path, target_label) in feature_in.items():
                encoder = TargetEncoder(cols=[feature], min_samples_leaf=20, smoothing=10)
                encoder = get_encoder_np(encoder, dict_path)
                added_data[f"{feature}_TE"] = encoder.fit_transform(df[feature], df[target_label])[feature]
                save_encoder_np(encoder, dict_path)
            to_concat_df = pd.DataFrame.from_dict(data = added_data)
            df = pd.concat([df, to_concat_df], axis = 1)
            return df
        return encode
    
class CountEncodeOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.feature_in = self.op.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = True
    
    def get_function_pd(self):
        from category_encoders.count import CountEncoder
        feature_in = copy.deepcopy(self.feature_in)
        def encode(df):
            added_data = {}
            for feature in feature_in:
                dict_path = None
                encoder = CountEncoder(cols=[feature])
                encoder = get_encoder_np(encoder, dict_path)
                added_data[f"{feature}_CE"] = encoder.fit_transform(df[feature])[feature]
                save_encoder_np(encoder, dict_path)
            to_concat_df = pd.DataFrame.from_dict(data = added_data)
            df = pd.concat([df, to_concat_df], axis = 1)
            return df
        return encode


from .base import BaseFeatureGenerator as super_class
from .featuretools_adaptor import FeaturetoolsBasedFeatureGenerator
from pyrecdp.core import SeriesSchema
from pyrecdp.primitives.operations import Operation
import pandas as pd
import numpy as np
from pandas.api import types as pdt
import copy
from featuretools.primitives.base import TransformPrimitive

class TypeCheckFeatureGenerator(super_class):
    def __init__(self, final = False, **kwargs):
        super().__init__(**kwargs)
   
    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        config = {}
        for idx in range(len(pa_schema)):
            pa_field = pa_schema[idx]
            if pa_field.is_categorical_and_string:
                pa_schema[idx] = SeriesSchema(pa_field.name, pd.StringDtype())
                config[pa_field.name] = 'str'
            elif pa_field.is_categorical:
                pa_schema[idx] = SeriesSchema(pa_field.name, pd.Int32Dtype())
                config[pa_field.name] = 'int'
        
        cur_idx = max_idx
        #pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'type_check', config = config)
        return pipeline, cur_idx, cur_idx

class IntTransformer(TransformPrimitive):
    name = "astype_int"
    return_type = int

    def get_function(self):
        def astype_int(array):
            try:
                ret = array.astype(int)
            except:
                ret = array
            return ret
        return astype_int

class FloatTransformer(TransformPrimitive):
    name = "astype_float"
    return_type = float

    def get_function(self):
        def astype_float(array):
            try:
                ret = array.astype(float)
            except:
                ret = array
            return ret
        return astype_float

class TypeConvertFeatureGenerator(FeaturetoolsBasedFeatureGenerator):
    def __init__(self, final = False, **kwargs):
        super().__init__(**kwargs)

    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        feature_in = {}
        feature_in_out_map = {}
        is_useful = False
        for pa_field in pa_schema:
            if pa_field.is_integer:
                op_class = IntTransformer
                feature_in[pa_field.name] = op_class
            if pa_field.is_float:
                op_class = FloatTransformer
                feature_in[pa_field.name] = op_class

        for in_feat_name, op_clz in feature_in.items():
            is_useful = True
            feature_in_out_map[in_feat_name] = []
            feature_in_out_map[in_feat_name].append((in_feat_name, op_clz))
        if is_useful:
            cur_idx = max_idx + 1
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = "astype", config = feature_in_out_map)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx
